/*
 * mod_suid.c for Apache 1.3.12
 *
 * Copyright 2002 by Igmar Palsenberg. Original idea by Anthony Howe.
 * Module skeleton code (c) Anthony Howe
 * Code stolen (c) Anthony Howe
 *
 *
 * See the LICENSE file for details about licensing issues
 *
 *
 * W A R N I N G
 * -------------
 *
 * USE OF THIS MODULE MAY OPEN SECURITY EXPLOITS WITHIN APACHE AND ITS
 * INSTALLED MODULES. THE AUTHOR DOES NOT RECOMMEND THE USE OF THIS MODULE.
 * THE AUTHOR DOES NOT ACCEPT ANY RESPONSIBLITY FOR SECURITY BREACHES ON
 * ANY WEB SERVER WHERE THIS MODULE IS EMPLOYED, NOR ANY SECURITY BREACHES
 * ON THOSE COMPUTERS ACCESSIBLE FROM A WEB SERVER EMPLOYING THIS MODULE.
 * USE AT YOUR OWN RISK.
 *
 * This module was created as a courtesy for Nadine Miller
 * <nmiller@cisco.com> upon request to the SAGE mailing list.
 *
 * This is a derived work from mod_become, however, allmost all big issues
 * of the original module have been fixed.
 *
 * SECURITY ISSUES
 * ---------------
 *
 *  1.	Apache 1.3.12 must be compiled with -DBIG_SECURITY_HOLE in order for
 *	this module to work.
 *
 *  2.	"User root" directive must be set so that the child processes using
 *	this module can perform a setuid() and setgid(). If the User
 *	directive is set to someone else then this module is disabled.
 *
 *  3.	Nothing is done about open file descriptors and therefore they could
 *	be subverted in some manner, such as to corrupt web server logs.
 *
 *  4.	Apache processes handlers in reverse order from which they were
 *	loaded. Its highly recommended that mod_suid be the last module in
 *	the LoadModule sequence, because in this way mod_suid's handler
 *	will be the first executed within its applied phase, so no other
 *	module can skip mod_suid.
 *
 *  6.	mod_suid is applied in phase 4, host acccess control, because this
 *	phase loosely corresponds with mod_suid's purpose, application of
 *	user/group file access control.
 *
 */
#define MODULE		"mod_suid"
#define AUTHOR		"Igmar Palsenberg igmar@jdimedia.nl"
#define VERSION		"1.1"
#define UID_MIN		50
#define GID_MIN		50

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_config.h"

#include <pwd.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifndef TRUE
# define TRUE		1
#endif

#ifndef FALSE
# define FALSE		0
#endif

/* Inline related stuff */
#ifdef INLINE 
# define MSINLINE inline
#else
# define MSINLINE
#endif

#define UNSET			-1
#define MERGE(p, c)		(c == UNSET ? p : c)

module MODULE_VAR_EXPORT suid_module;

#define POLICY_OFF				0
#define POLICY_FILE				1	/* Not recommended, very loose controls. */
#define POLICY_USER_GROUP		2	/* Default behaviour like User & Group. */
#define POLICY_DOCUMENT_ROOT	3	/* Fairly strict. */
#define POLICY_PARENT_DIRECTORY	4	/* Best choice. */

static const char *def_users[] = {
	"wwwuser", "httpd", "www", "web", "nobody", (char *) 0
};

/* Prototypes */
static MSINLINE int change_to_user_group(request_rec * r, uid_t uid, gid_t gid);
static MSINLINE int change_to_root(request_rec * r);
static MSINLINE int change_to_apache_user(request_rec * r);
static int istrue(char *s);
static void * create_dir_entry(pool *p, char *dir);
static void * create_server_entry(pool *p, server_rec *s);
static void init_module(server_rec *s, pool *p);
static void * merge_dir_entries(pool *p, void * d_base, void * d_child);
static void * merge_server_entries(pool *p, void * parent, void * current);
static void child_init(server_rec *s, pool *p);
static void child_exit(server_rec * s, pool * p);
static const char * suid(cmd_parms *cmd, void *dconfig, char *one, char *two);
static const char * suid_policy(cmd_parms *cmd, void *dconfig, char *one);
static const char * suid_enable(cmd_parms *cmd, void *dconfig, char *one);
static int uri_handler(request_rec * r);
static int access_handler(request_rec *r);
static int fixup_handler(request_rec *r);

/*
 * Per server, host, and directory information.
 */
typedef struct config {
	int allow_root;
	int policy;
	uid_t uid;
	gid_t gid;
} t_config;

static const char stype[] = "Server";
static const char dtype[] = "Directory";

static MSINLINE int change_to_user_group(request_rec * r, uid_t uid, gid_t gid)
{
	if (setresgid(gid, gid, 0) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_user_group : setresgid(%d,%d,0) failed (\"%s\")", ap_get_server_name(r), r->the_request, gid, gid, strerror(errno));
		}
		return FALSE;
	}
	if (setresuid(uid, uid, 0) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_user_group : setresuid(%d,%d,0) failed (\"%s\")", ap_get_server_name(r), r->the_request, uid, uid, strerror(errno));
		}
		return FALSE;
	}
	
	return TRUE;
}

static MSINLINE int change_to_root(request_rec * r)
{
	if (setreuid(-1,0) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_root : setreuid(-1,0) failed (\"%s\")", ap_get_server_name(r), r->the_request, strerror(errno));
		}
		return FALSE;
	}
	if (setreuid(0,-1) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_root : setreuid(0,-1) failed (\"%s\")", ap_get_server_name(r), r->the_request, strerror(errno));
		}
		return FALSE;
	}
	if (setregid(-1,0) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_root : setregid(-1,0) failed (\"%s\")", ap_get_server_name(r), r->the_request, strerror(errno));
		}
		return FALSE;
	}
	if (setregid(0,-1) != 0) {
		if (r) {
			ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: change_to_root : setregid(0,-1) failed (\"%s\")", ap_get_server_name(r), r->the_request, strerror(errno));
		}
		return FALSE;
	}
	
	return TRUE;
}

static MSINLINE int change_to_apache_user(request_rec * r)
{
	const char **list;
	struct passwd * pw;

	/* Find Apache user */
	for (list = def_users; *list != (char *) 0; ++list) {
		if ((pw = getpwnam(*list)) != (struct passwd *) 0) {
			break;
		}
	}
	
	if (pw == NULL || pw->pw_uid == 0 || pw->pw_gid == 0)
		return FALSE;
	
	/* Change back to root.. Else it's a No Go */
	if (change_to_root(r) == FALSE)
		return FALSE;
	if (change_to_user_group(r, pw->pw_uid, pw->pw_gid) == FALSE)
		return FALSE;
		
	return TRUE;
}

static int istrue(char *s)
{
	if (ap_strcasecmp_match(s, "enable") == 0)	/* disable */
		return TRUE;
	if (ap_strcasecmp_match(s, "true") == 0)	/* false */
		return TRUE;
	if (ap_strcasecmp_match(s, "yes") == 0)		/* no */
		return TRUE;
	if (ap_strcasecmp_match(s, "set") == 0)		/* reset */
		return TRUE;
	if (ap_strcasecmp_match(s, "ok") == 0)		/* bogus */
		return TRUE;
	if (ap_strcasecmp_match(s, "on") == 0)		/* off */
		return TRUE;
	if (ap_strcasecmp_match(s, "1") == 0)		/* 0 */
		return TRUE;

	return FALSE;
}

static void * create_dir_entry(pool *p, char *dir)
{
	t_config *config = (t_config *) ap_pcalloc(p, sizeof(t_config));

	/* This option always false and must be explicitly set true. */
	config->allow_root = 0;

	/* These values can be inherited from the parent server. */
	config->policy = UNSET;
	config->uid = UNSET;
	config->gid = UNSET;

	return (void *) config;
}

/*
 * This module is called once _before_ logging is enabled and
 * the first call to init_module() to create the parent server
 * entry.  After logging is started and init_module(), this
 * is called for each virtual host where a directive related
 * to this module exhists.
 */
static void * create_server_entry(pool *p, server_rec *s)
{
	t_config *config;

	config = (t_config *) create_dir_entry(p, (char *) 0);

	return (void *) config;
}

static void init_module(server_rec *s, pool *p)
{
	struct group *gr;
	struct passwd *pw;
	const char **list;

	t_config *config = (t_config *) ap_get_module_config(
		s->module_config, &suid_module
	);

	ap_add_version_component("mod_suid/"VERSION);

	/* Look for some of the more common web server account names to
	 * use as the initial defaults just in case the admin. forgets
	 * to add DoSuid directives.  If nothing is found then the default
	 * will be UNSET, which will generate a runtime error if left that
	 * way.
	 *
	 * The checks for UNSET here are because, init_module() is called
	 * twice, once before all the virtual hosts are configured and once
	 * after. The O'Reilly book "Writing Apache Modules" claims this is
	 * to test that the module can handle a restart, but I think this is
	 * incorrect.  Its probably to allow server initialisation to be
	 * influenced by virtual hosts if required.
	 */
	if (config->uid == UNSET) {
		for (list = def_users; *list != (char *) 0; ++list) {
			if ((pw = getpwnam(*list)) != (struct passwd *) 0) {
				config->uid = pw->pw_uid;
				break;
			}
		}
	}

	if (config->gid == UNSET) {
		for (list = def_users; *list != (char *) 0; ++list) {
			if ((gr = getgrnam(*list)) != (struct group *) 0) {
				config->gid = gr->gr_gid;
				break;
			}
		}
	}

	if (config->policy == UNSET)
		config->policy = POLICY_USER_GROUP;
}

static void * merge_dir_entries(pool *p, void * d_base, void * d_child)
{
	t_config *child = (t_config *) d_child;
	t_config *base = (t_config *) d_base;
	t_config *merged;

	if (child == (void *) 0)
		return base;
	if (base == (void *) 0)
		return child;

	merged = (t_config *) ap_pcalloc(p, sizeof (t_config));

	merged->allow_root = child->allow_root | base->allow_root;

	merged->policy = MERGE(base->policy, child->policy);
	merged->uid = MERGE(base->uid, child->uid);
	merged->gid = MERGE(base->gid, child->gid);

	return (void *) merged;
}

static void * merge_server_entries(pool *p, void * parent, void * current)
{
	return merge_dir_entries(p, parent, current);
}

static void child_init(server_rec *s, pool *p)
{
	change_to_apache_user(NULL);
}

static void child_exit(server_rec * s, pool * p)
{
	/*
     * No idea if this really has some use
     * We do it anyway :)
	*/
	/* FIXME */
	/* change_to_root(NULL); */
}

/*
 * Become user {id}
 * Become group {id}
 *
 * Context: server, virtual host, <directory>, <location>
 */
static const char * suid(cmd_parms *cmd, void *dconfig, char *one, char *two)
{
	t_config *config;
	const char *what, *where;

	/* Is it a server configuration? */
	if (cmd->path == (char *) 0 || *cmd->path == '\0') {
		what = stype;
		where = cmd->server->server_hostname;
		config = (t_config *) ap_get_module_config(
			cmd->server->module_config, &suid_module
		);
	} else {
		what = dtype;
		where = cmd->path;
		config = (t_config *) dconfig;
	}

	if (config == (t_config *) 0)
		return (const char *) 0;

	if (ap_strcasecmp_match(one, "user") == 0) {
		config->uid = ap_uname2id(two);
	} else if (ap_strcasecmp_match(one, "group") == 0) {
		config->gid = ap_gname2id(two);
	} else {
		return "Invalid argument.";
	}

	return (const char *) 0;
}

/*
 * SuidPolicy {policy}
 *
 * where policy is one of:
 *
 *	file, user-group, document-root, parent-directory
 *
 * Context: server, virtual host, <directory>, <location>
 */
static const char * suid_policy(cmd_parms *cmd, void *dconfig, char *one)
{
	t_config *config;
	const char *what, *where;

	if (cmd->path == (char *) 0 || *cmd->path == '\0') {
		what = stype;
		where = cmd->server->server_hostname;
		config = (t_config *) ap_get_module_config(
			cmd->server->module_config, &suid_module
		);
	} else {
		what = dtype;
		where = cmd->path;
		config = (t_config *) dconfig;
	}

	if (config == (t_config *) 0)
		return (const char *) 0;

	if (ap_strcasecmp_match(one, "file") == 0) {
		config->policy = POLICY_FILE;
	} else if (ap_strcasecmp_match(one, "user*group") == 0) {
		config->policy = POLICY_USER_GROUP;
	} else if (ap_strcasecmp_match(one, "doc*root") == 0) {
		config->policy = POLICY_DOCUMENT_ROOT;
	} else if (ap_strcasecmp_match(one, "parent*dir*") == 0) {
		config->policy = POLICY_PARENT_DIRECTORY;
	} else {
		return "Invalid policy.";
	}

	return (const char *) 0;
}

/*
 * SuidEnable {boolean}
 *
 * Context: server, virtual host, <directory>, <location>
 */
static const char * suid_enable(cmd_parms *cmd, void *dconfig, char *one)
{
	t_config *config;
	const char *what, *where;

	/* Is it a server configuration? */
	if (cmd->path == (char *) 0 || *cmd->path == '\0') {
		what = stype;
		where = cmd->server->server_hostname;
		config = (t_config *) ap_get_module_config(
			cmd->server->module_config, &suid_module
		);
	} else {
		what = dtype;
		where = cmd->path;
		config = (t_config *) dconfig;
	}

	if (config != (t_config *) 0)
		config->allow_root = istrue(one);

	return (const char *) 0;
}

static int uri_handler(request_rec * r)
{
	t_config *config, *dconf, *sconf;
	
	dconf = (t_config *) ap_get_module_config(
		r->per_dir_config, &suid_module
	);

	sconf = (t_config *) ap_get_module_config(
		r->server->module_config, &suid_module
	);
	
	config = (t_config *) merge_dir_entries(r->pool, sconf, dconf);
	
	/* Change to root. We need the privileges */
	if (!config->allow_root) {
		change_to_apache_user(r); 
	} else {
		/* Change to root. We need the privileges */
		change_to_root(r); 
	}

	return DECLINED;
}

static int access_handler(request_rec *r)
{
	uid_t this_uid;
	gid_t this_gid;
	struct stat linfo;
	struct passwd *pw, *apache;
	const char *dir = (const char *) 0;
	t_config *config, *dconf, *sconf;
	const char **list;

	dconf = (t_config *) ap_get_module_config(
		r->per_dir_config, &suid_module
	);

	sconf = (t_config *) ap_get_module_config(
		r->server->module_config, &suid_module
	);

	/* Server and directory configuration structures are the same,
	 * but the information is not merged by Apache.  Do it ourselves.
	 */
	config = (t_config *) merge_dir_entries(r->pool, sconf, dconf);

	if (!config->allow_root)
		return DECLINED;
	
	if (getuid()) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: not root but (\"%d\")", ap_get_server_name(r), r->the_request, getuid());
		return HTTP_SERVICE_UNAVAILABLE;
	}
	/*
     * Skip this for CGI stuff
    */

	if (config == (t_config *) 0) {
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "%s %s: No user, group, or policy set", ap_get_server_name(r), r->the_request);
		return HTTP_SERVICE_UNAVAILABLE;
	}

#ifndef IGNORE_SYMLINK_OWNER
	/* Precaution: make sure that the owner of a symbolic link is the
	 * same as that of the file the link references. Call me paranoid.
	 * I could have tested further down just before the setgid(), but
	 * I figured it has to get done may as well be sooner than later.
	 */
	if (lstat(r->filename, &linfo) == 0 && S_ISLNK(linfo.st_mode)) {
		if (r->finfo.st_uid != linfo.st_uid) {
			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "Symbolic link \"%s\" to target file not owned by same user.", r->filename);
			return HTTP_PRECONDITION_FAILED;
		}

		if (r->finfo.st_gid != linfo.st_gid) {
			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "Symbolic link \"%s\" to target file not owned by same group.", r->filename);
			return HTTP_PRECONDITION_FAILED;
		}
	}
#endif

	switch (config->policy) {
	default:
	case POLICY_USER_GROUP:
		/* Use the user & group specified by Become directive. */
		this_uid = config->uid;
		this_gid = config->gid;
		break;
	case POLICY_FILE:
		/* Use the user & group of the requested file. */
		this_uid = r->finfo.st_uid;
		this_gid = r->finfo.st_gid;
		break;
	case POLICY_DOCUMENT_ROOT:
		/* Use the user & group of the site's document root. */
		if (r->uri[0] == '/' && r->uri[1] == '~' && isalnum(r->uri[2])) {
			/* Make sub-request here to find out user's private
			 * document root for their web site, which could have
			 * a different group id from their account profile
			 * and/or home directory.
			 */
			int status;
			request_rec *sub;
			const char *user;
			const char *remainder = r->uri + 2;

			user = ap_getword(r->pool, &remainder, '/');
			user = ap_pstrcat(r->pool, "/~", user, "/", NULL);

			sub = ap_sub_req_lookup_uri(user, r);

			dir = ap_pstrdup(r->pool, sub->filename);
			this_uid = sub->finfo.st_uid;
			this_gid = sub->finfo.st_gid;
			status = sub->status;

			ap_destroy_sub_req(sub);

			/* When the subrequest to find the ~user's document
			 * root fails, it might be that the user or their
			 * document root has been removed.  We could fall
			 * back on the default user & group, but I think
			 * an error is better.
			 *
			 * NOTE that the 300 class of errors concerning moves
			 * and redirects could be handled differently.
			 */
			if (!ap_is_HTTP_SUCCESS(status)) {
				ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "%s %s: Cannot find \"~%s\" document root", ap_get_server_name(r), r->the_request, user);
				return HTTP_FORBIDDEN;
			}
		} else {
			dir = ap_document_root(r);
		}
		break;
	case POLICY_PARENT_DIRECTORY:
		/* Use the user & group of the immediate parent directory of
		 * our request. If the request refers to a directory, then
		 * the user & group of the directory itself is used.
		 */
		if (S_ISDIR(r->finfo.st_mode)) {
			/* Directory request considered as its own parent. */
			this_uid = r->finfo.st_uid;
			this_gid = r->finfo.st_gid;
		} else {
			/* Parent directory of our request. */
			dir = ap_make_dirstr_parent(r->pool, r->filename);
		}
		break;
	}

	if (dir != (const char *) 0) {
		struct stat dinfo;

		/* We're still user root at this point, so if we can't
		 * stat() a directory, most likely it doesn't exist,
		 * since we have sufficent permissions to look at it.
		 */
		if (stat(dir, &dinfo) != 0) {
			ap_log_rerror( APLOG_MARK, APLOG_ERR, r, "%s %s: Cannot stat(%s)", ap_get_server_name(r), r->the_request, dir);
			return HTTP_FORBIDDEN;
		}

		this_uid = dinfo.st_uid;
		this_gid = dinfo.st_gid;
	}

	ap_log_rerror( APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r, "%s %s (%ld): policy=%d, target uid=%ld, gid=%ld, allow_root=%d, file=%s",
		ap_get_server_name(r), r->the_request, (long) getpid(),
		config->policy, (long) this_uid, (long) this_gid,
		config->allow_root, r->filename
	);

	if (this_uid == UNSET) {
		ap_log_rerror(
			APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
			"%s %s: No user ID set",
			ap_get_server_name(r), r->the_request
		);
		return HTTP_SERVICE_UNAVAILABLE;
	}

	if (this_gid == UNSET) {
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "%s %s: No group ID set", ap_get_server_name(r), r->the_request);
		return HTTP_SERVICE_UNAVAILABLE;
	}

	if (this_gid == 0 && !config->allow_root) {
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "%s %s: Not allowed to suid to group root", ap_get_server_name(r), r->the_request);
		return HTTP_FORBIDDEN;
	}

	/* Failure to getpwuid() could mean /etc/passwd has changed while
	 * we are processing this request.  The error is logged, instead
	 * of attempting to fallback on the default Become user.
	 */
	if ((pw = getpwuid(this_uid)) == (struct passwd *) 0) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: Cannot getpwuid(%ld)", ap_get_server_name(r), r->the_request, (long) this_uid);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	/* Failure of initgroups() and setgroups() could mean /etc/passwd
	 * and/or /etc/group has changed while we are processing this request.
	 * The error is logged, instead of attempting to fallback on the
	 * default Become user & group.
	 */
	if (initgroups(pw->pw_name, this_gid) < 0) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: initgroups(\"%s\", %ld)", ap_get_server_name(r), r->the_request, pw->pw_name, (long) this_gid);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	if (this_uid == 0 && !config->allow_root) {
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "%s %s: Not allowed to suid to user root", ap_get_server_name(r), r->the_request);
		return HTTP_FORBIDDEN;
	}

	/* We don't want a switch to UID or GID below a certain treshold */
	if (this_uid <= UID_MIN) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: UID %ld, must be above %ld",
		ap_get_server_name(r), r->the_request, (long) this_uid, (long) UID_MIN);
		return HTTP_FORBIDDEN;
	}
	if (this_gid <= GID_MIN) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s %s: GID %ld, must be above %ld", ap_get_server_name(r), r->the_request, (long) this_gid, (long) GID_MIN);
		return HTTP_FORBIDDEN;
	}

	change_to_user_group(r, this_uid, this_gid);

	/* Success, continue remainder of handler chain. */
	return DECLINED;
}

static int fixup_handler(request_rec *r)
{
	t_config *config, *dconf, *sconf;
	
	dconf = (t_config *) ap_get_module_config(
		r->per_dir_config, &suid_module
	);

	sconf = (t_config *) ap_get_module_config(
		r->server->module_config, &suid_module
	);
	
	config = (t_config *) merge_dir_entries(r->pool, sconf, dconf);
	
	if (config->allow_root)
			change_to_apache_user(r);

	return DECLINED;
}

command_rec command_table[] = {
	{ "Suid", suid, NULL, RSRC_CONF|ACCESS_CONF, TAKE2,
	  "Suid user {id}\nSuid group {id}\n" },

	{ "SuidPolicy", suid_policy, NULL, RSRC_CONF|ACCESS_CONF, TAKE1,
	  "SuidPolicy {policy}\n where policy is one of:\n"
	  "\tfile, user-group, document-root, parent-directory\n" },

	{ "SuidEnable", suid_enable, NULL, RSRC_CONF|ACCESS_CONF, TAKE1,
	  "SuidRoot {boolean}\nAllow the server or virtual host to be root user or group.\n" },

	{ NULL }
};


/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT suid_module = {
	STANDARD_MODULE_STUFF,
	init_module,			/* module initializer                  */
	create_dir_entry,		/* create per-dir    config structures */
	merge_dir_entries,		/* merge  per-dir    config structures */
	create_server_entry,	/* create per-server config structures */
	merge_server_entries,	/* merge  per-server config structures */
	command_table,			/* table of config file commands       */
	NULL,					/* [#8] MIME-typed-dispatched handlers */
	uri_handler,			/* [#1] URI to filename translation    */
	NULL,					/* [#4] validate user id from request  */
	NULL,					/* [#5] check if the user is ok _here_ */
	access_handler,			/* [#3] check access by host address   */
	NULL,					/* [#6] determine MIME type            */
	NULL,					/* [#7] pre-run fixups                 */
	fixup_handler,			/* [#9] log a transaction              */
	NULL,					/* [#2] header parser                  */
	child_init,				/* child_init					       */
	child_exit,				/* child_exit			       		   */
	NULL					/* [#0] post read-request              */
};
